import psycopg2
import configparser
import requests
import time
import csv
from datetime import datetime
import traceback
import sys
import getopt
import os

#
# Converts time value to seconds.
#
def to_seconds(str):
    
    if str is None:
        return None
    
    wait_time = str.split(":")
    
    if (len(wait_time) == 1):
        return None
    
    counter = 0
    seconds  = 0
    
    for t in wait_time:
        t = int(t)
        if counter == 0:
            seconds += (3600 * t)
        elif counter == 1:
            seconds += (60 * t)
        else:
            seconds += t
        counter += 1    
  
    return seconds

#
#  Converts CSV boolean to postgres boolean
#     
def convert_boolean(val):
    
    new_val = None
    if val:
        if val == 'FALSE':
            return False
        elif val == 'TRUE':
            return True
    return new_val

#
# Converst CSV blank or empty string to postgres null
#
def convert_null(val):
    if val == '':
        return None
    elif val is None:
        return None
    else:
        return val

#
# Checks if column exists in file being processed.
# If not, returns None so it's not inluded in insert stmt
#         
def check_if_column_exists(columns, column, row):
    lookup_column = column.lower()
    if lookup_column in columns:
        return row[column]
    else:
        return None

#
# Convert csv int data to postgress int
#
def convert_int(num):
    if num is None:
        return None
    elif num == '':
        return None
    else:
        return int(float(num))

#
# Remove extra characters not needed. 
# For example, $10.00 -> 10.00
#         
def strip_if_not_null (val, char):
    if val is None:
        return None
    else:
        return val.strip(char)

# 
# If visit id exists in visits.id, return true, else false
#     
def record_exists(mydb, cursor, row_id):
    cursor.execute("select id from visits where visit_id = '" + row_id + "'" )
    visit_id = cursor.fetchone()
    if visit_id is not None:
        return True
    else:
        return False

#
# Updates visit table with passed in values.
#
def update_visit(mydb, cursor, vals, visit_id, rownum):
    
    # append visit_id to end of tuple for where clause
    id_tuple = (visit_id,)
    new_vals = vals + id_tuple
    
    sql = """update visits 
                set visit_id = %s,
                    visit_type = %s,
                    patient_device = %s,
                    patient_first_name = %s,
                    patient_last_name = %s,
                    patient_dob = %s, 
                    patient_gender = %s,
                    patient_relation_to_account = %s,
                    account_first_name = %s,
                    account_last_name = %s,
                    account_dob = %s,
                    account_gender = %s,
                    assigned_provider = %s,
                    actual_provider = %s,
                    epic_Instance = %s,
                    epic_department = %s,
                    eid = %s,
                    epic_timeslot = %s,
                    current_state = %s,
                    patient_email = %s,
                    patient_phone = %s,
                    intake_billing_address = %s,
                    epic_billing_address = %s,
                    successful = %s,
                    payment_type = %s,
                    insurance_id = %s,
                    patient_subscriber_relationship = %s,
                    subscriber_member_id = %s,
                    subscriber_first_name = %s,
                    subscriber_last_name = %s,
                    subscriber_dob = %s, 
                    transaction_id = %s,
                    amount_charged = %s,
                    waived = %s,
                    waive_reason = %s,
                    status = %s,
                    ready_for_visit = %s,
                    high_priority_visit = %s,
                    wait_time_in_seconds = %s,
                    visit_duration_in_seconds = %s,
                    requested_date = %s, 
                    requested_time = %s,
                    waiting_room_date = %s,
                    waiting_room_time = %s,
                    in_visit_date = %s,
                    in_visit_time = %s,
                    done_date = %s,
                    done_time = %s,
                    declined_date = %s,
                    declined_time = %s,
                    declined_reason = %s,
                    decliner_name = %s,
                    decliner_roles = %s,
                    cancelled_date = %s,
                    cancelled_time = %s,
                    reason_for_visit = %s,
                    waiting_room_id = %s,
                    conference_room_id = %s,
                    app_brand_name = %s,
                    app_version = %s,
                    device_os_version = %s,
                    accept_tou = %s,
                    technology_feedback = %s,
                    service_key = %s,
                    client_id = %s,
                    discount_amount = %s,
                    initial_estimated_wait_time_minutes = %s,
                    initial_estimated_wait_time_verbiage = %s,
                    address_flagged = %s,
                    visit_rating = %s,
                    contact_authorization = %s,
                    additional_feedback = %s,
                    provider_feedback = %s,
                    customer_name = %s,
                    practice_id = %s,
                    practice_name = %s,
                    practice_region_id = %s,
                    practice_region_name = %s,
                    removed_from_rcpq = %s, 
                    time_spent_in_rcpq = %s, 
                    payment_module_complete = %s,
                    icc_photos_uploaded = %s, 
                    time_to_visit = %s, 
                    visit_in_rcpq = %s, 
                    doc_signing_complete = %s,
                    source_file = %s,
                    last_update_date = now()
              where visit_id = %s
                """
    try:
        cursor.execute(sql, new_vals)
        mydb.commit()
    except Exception as e:
        print("Row " + str(rownum) + " skipped due to error:")
        traceback.print_exception(*sys.exc_info())
        mydb.rollback() 

#
# Inserts new visits record with passed in values.
#
def insert_visit(mydb, cursor, vals, rownum):
    sql = """INSERT 
               INTO visits (visit_id,
                 visit_type,
                 patient_device,
                 patient_first_name,
                 patient_last_name,
                 patient_dob, 
                 patient_gender,
                 patient_relation_to_account,
                 account_first_name,
                 account_last_name,
                 account_dob,
                 account_gender,
                 assigned_provider,
                 actual_provider,
                 epic_Instance,
                 epic_department,
                 eid,
                 epic_timeslot,
                 current_state,
                 patient_email,
                 patient_phone,
                 intake_billing_address,
                 epic_billing_address,
                 successful,
                 payment_type,
                 insurance_id ,
                 patient_subscriber_relationship,
                 subscriber_member_id,
                 subscriber_first_name,
                 subscriber_last_name,
                 subscriber_dob, 
                 transaction_id,
                 amount_charged,
                 waived,
                 waive_reason,
                 status,
                 ready_for_visit,
                 high_priority_visit,
                 wait_time_in_seconds,
                 visit_duration_in_seconds,
                 requested_date, 
                 requested_time,
                 waiting_room_date,
                 waiting_room_time,
                 in_visit_date,
                 in_visit_time,
                 done_date,
                 done_time,
                 declined_date,
                 declined_time,
                 declined_reason,
                 decliner_name,
                 decliner_roles,
                 cancelled_date,
                 cancelled_time,
                 reason_for_visit,
                 waiting_room_id,
                 conference_room_id,
                 app_brand_name,
                 app_version,
                 device_os_version,
                 accept_tou,
                 technology_feedback,
                 service_key,
                 client_id,
                 discount_amount,
                 initial_estimated_wait_time_minutes,
                 initial_estimated_wait_time_verbiage,
                 address_flagged,
                 visit_rating,
                 contact_authorization,
                 additional_feedback,
                 provider_feedback,
                 customer_name,
                 practice_id,
                 practice_name,
                 practice_region_id,
                 practice_region_name,
                 removed_from_rcpq, 
                 time_spent_in_rcpq, 
                 payment_module_complete,
                 icc_photos_uploaded, 
                 time_to_visit, 
                 visit_in_rcpq, 
                 doc_signing_complete,
                 source_file,
                 creation_date,
                 last_update_date) 
         VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                 %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                 %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                 %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, now(), now())"""

    try:
        cursor.execute(sql, vals)
        mydb.commit()
    except Exception as e:
        print("Row " + str(rownum) + " skipped due to error:")
        print(cursor.mogrify(sql, vals))
        traceback.print_exception(*sys.exc_info())
        mydb.rollback()

#
# Converts passed in visits CSV to postgress friendly values and checks
# to see what columns are passed in to make correct insert/update
# statment.
#         
def insert_csv_into_database(mydb, in_file):
    print("Inserting file:", in_file)
    
    cursor = mydb.cursor()

    file_name = os.path.basename(in_file)

    # Open file and iterate to get total rows
    totalrows = sum(1 for _ in open(in_file))
    
    # Open again to process each row from start
    csv_data = csv.DictReader(open(in_file))

    # Lower case heading to avoid inconsistencies when running
    # from API (mixed case) vs backfills from Athena (lower case)
    columns = [name.lower() for name in csv_data.fieldnames]

    if totalrows <= 1:
        print("No data found to insert.")
        return

    row_number=2
    for row in csv_data:
        
        # Creates correct insert/update tuple for values passed. If value is not passed
        # for a field, a null will be inserted into database for that field.
        vals = (convert_null(check_if_column_exists(columns, "id", row)),
                convert_null(check_if_column_exists(columns, "Visit_Type", row)),
                convert_null(check_if_column_exists(columns, "Patient_Device", row)),
                convert_null(check_if_column_exists(columns, "Patient_First_Name", row)),
                convert_null(check_if_column_exists(columns, "Patient_Last_Name", row)),
                convert_null(check_if_column_exists(columns, "Patient_DOB", row)), 
                convert_null(check_if_column_exists(columns, "Patient_Gender", row)),
                convert_null(check_if_column_exists(columns, "Patient_Relation_to_Account", row)),
                convert_null(check_if_column_exists(columns, "Account_First_Name", row)),
                convert_null(check_if_column_exists(columns, "Account_Last_Name", row)),
                convert_null(check_if_column_exists(columns, "Account_DOB", row)),
                convert_null(check_if_column_exists(columns, "Account_Gender", row)),
                convert_null(check_if_column_exists(columns, "Assigned_Provider", row)),
                convert_null(check_if_column_exists(columns, "Actual_Provider", row)),
                convert_null(check_if_column_exists(columns, "Epic_Instance", row)),
                convert_null(check_if_column_exists(columns, "Epic_Department", row)),
                convert_null(check_if_column_exists(columns, "EID", row)),
                convert_null(check_if_column_exists(columns, "Epic_Timeslot", row)),
                convert_null(check_if_column_exists(columns, "Current_State", row)),
                convert_null(check_if_column_exists(columns, "Patient_Email", row)),
                convert_null(check_if_column_exists(columns, "Patient_Phone", row)),
                convert_null(check_if_column_exists(columns, "Intake_Billing_Address", row)),
                convert_null(check_if_column_exists(columns, "Epic_Billing_Address", row)),
                convert_boolean(check_if_column_exists(columns, "Successful", row)),
                convert_null(check_if_column_exists(columns, "Payment_Type", row)),
                convert_null(check_if_column_exists(columns, "Insurance_ID", row)),
                convert_null(check_if_column_exists(columns, "Patient_Subscriber_Relationship", row)),
                convert_null(check_if_column_exists(columns, "Subscriber_Member_ID", row)),
                convert_null(check_if_column_exists(columns, "Subscriber_First_Name", row)),
                convert_null(check_if_column_exists(columns, "Subscriber_Last_Name", row)),
                convert_null(check_if_column_exists(columns, "Subscriber_DOB", row)), 
                convert_null(check_if_column_exists(columns, "Transaction_ID", row)),
                convert_null(check_if_column_exists(columns, "Amount_Charged", row)),
                convert_boolean(check_if_column_exists(columns, "Waived", row)),
                convert_null(check_if_column_exists(columns, "Waive_Reason", row)),
                convert_null(check_if_column_exists(columns, "Status", row)),
                convert_boolean(check_if_column_exists(columns, "Ready_for_Visit", row)),
                convert_boolean(check_if_column_exists(columns, "High_Priority", row)),
                to_seconds(check_if_column_exists(columns, "Wait_Time", row)),
                to_seconds(check_if_column_exists(columns, "Visit_Duration", row)),
                convert_null(check_if_column_exists(columns, "Requested_Date", row)), 
                convert_null(check_if_column_exists(columns, "Requested_Time", row)),
                convert_null(check_if_column_exists(columns, "Waiting_Room_Date", row)),
                convert_null(check_if_column_exists(columns, "Waiting_Room_Time", row)),
                convert_null(check_if_column_exists(columns, "In_Visit_Date", row)),
                convert_null(check_if_column_exists(columns, "In_Visit_Time", row)),
                convert_null(check_if_column_exists(columns, "Done_Date", row)),
                convert_null(check_if_column_exists(columns, "Done_Time", row)),
                convert_null(check_if_column_exists(columns, "Declined_Date", row)),
                convert_null(check_if_column_exists(columns, "Declined_Time", row)),
                convert_null(check_if_column_exists(columns, "Declined_Reason", row)),
                convert_null(check_if_column_exists(columns, "Decliner_Name", row)),
                convert_null(check_if_column_exists(columns, "Decliner_Roles", row)),
                convert_null(check_if_column_exists(columns, "Cancelled_Date", row)),
                convert_null(check_if_column_exists(columns, "Cancelled_Time", row)),
                convert_null(check_if_column_exists(columns, "Reason_for_Visit", row)),
                convert_null(check_if_column_exists(columns, "Waiting_Room_ID", row)),
                convert_null(check_if_column_exists(columns, "Conference_Room_ID", row)),
                convert_null(check_if_column_exists(columns, "App_Brand_Name", row)),
                convert_null(check_if_column_exists(columns, "App_Version", row)),
                convert_null(check_if_column_exists(columns, "Device_OS_Version", row)),
                convert_boolean(check_if_column_exists(columns, "Accept_TOU", row)),
                convert_null(check_if_column_exists(columns, "Technology_Feedback", row)),
                convert_null(check_if_column_exists(columns, "Service_Key", row)),
                convert_null(check_if_column_exists(columns, "Client_ID", row)),
                convert_null(strip_if_not_null(check_if_column_exists(columns, "Discount_Amount", row),'$')),
                convert_null(convert_int(check_if_column_exists(columns, "Initial_Estimated_Wait_Time_Minutes", row))),
                convert_null(check_if_column_exists(columns, "Initial_Estimated_Wait_Time_Verbiage", row)),
                convert_boolean(check_if_column_exists(columns, "Address_Flagged", row)),
                convert_null(check_if_column_exists(columns, "Visit_Rating", row)),
                convert_boolean(check_if_column_exists(columns, "Contact_Authorization", row)),
                convert_null(check_if_column_exists(columns, "Additional_Feedback", row)),
                convert_null(check_if_column_exists(columns, "Provider_Feedback", row)),
                convert_null(check_if_column_exists(columns, "Customer_Name", row)),
                convert_null(check_if_column_exists(columns, "Practice_ID", row)),
                convert_null(check_if_column_exists(columns, "Practice_Name", row)),
                convert_null(check_if_column_exists(columns, "Practice_Region_ID", row)),
                convert_null(check_if_column_exists(columns, "Practice_Region_Name", row)), 
                convert_null(check_if_column_exists(columns, 'Removed_from_RCPQ',  row)), 
                convert_null(check_if_column_exists(columns, 'Time_Spent_in_RCPQ',  row)), 
                convert_boolean(check_if_column_exists(columns, 'Payment_Module_Complete',  row)), 
                convert_boolean(check_if_column_exists(columns, 'ICC_Photos_Uploaded',  row)), 
                convert_null(check_if_column_exists(columns, 'Time_to_Visit',  row)), 
                convert_boolean(check_if_column_exists(columns, 'Visit_in_RCPQ',  row)), 
                convert_boolean(check_if_column_exists(columns, 'Doc_Signing_Complete', row)),
                file_name)

        if record_exists(mydb, cursor, row["id"]):
            update_visit(mydb, cursor, vals, row["id"], row_number)
        else:
            insert_visit(mydb, cursor, vals, row_number)

        row_number = row_number + 1

            

            
        
    cursor.close()
    
#===================================================================================================

config = configparser.ConfigParser()
#
# config.ini has database and api credentials
# stored. This file must be created on each instance
# and made read only for owner only.  This file is
# checked into repo or included in docker image.
#
try:
    with open('config.ini') as f:
        config.read_file(f)
except IOError:
    print("config.ini not found.")
    quit()

mydb = psycopg2.connect(
  host=config['DATABASE']['host'],
  user=config['DATABASE']['username'],
  password=config['DATABASE']['password'],
  database=config['DATABASE']['database']
)

#
# download_directory is where the files downloaded from 
# API are stored. 
#
api = config['MISC']['api']
download_dir = config['MISC']['download_directory']
testing = False

now = datetime.now()
today = now.strftime("%Y%m%d_%H%M%S")

#
# Testing uses UAT API pulls
# TODO: Paramaterize this.
#
if testing:
    endpoint_section = 'UAT_END_POINTS'
    secret_section = 'UAT_SECRETS'
    file_suffix = 'uat'
else:
    endpoint_section = 'PROD_END_POINTS'
    secret_section = 'PROD_SECRETS'
    file_suffix = 'prod'


for (each_key, each_val) in config.items(endpoint_section):
    
    customer = each_key
    end_point = each_val

    for (secret_key, secret_val) in config.items(secret_section):
        if secret_key == customer:
            secret = secret_val


report_types = ['virtual','retail']
report_ids = {}

headers = {
'Authorization' : 'dc-token ' + secret,
'Content-type': 'application/json',
}

# Submit reports
for report_type in report_types:

    data = "{{\"type\":\"{}\"}}".format(report_type)
    response = requests.post('https://' + end_point + api, headers=headers, data=data)

    if response.status_code == 200:
        report_ids[report_type]=response.json()['id']
    else:
        if 'message' in response.json():
            msg = response.json()['message']
        elif 'error' in response.json():
            msg = response.json()['error']
        else:   
            print("Unknown Format Returned:", response.json())

        print(response.status_code,"Error submitting report",msg)
        # TODO: Send email or alert

# Wait for reports to finish
for report_id in report_ids:
    while True:
        response = requests.get('https://' + end_point + api + '/' + report_ids[report_id], headers=headers)
    
        print(report_ids[report_id],'->',response.json()['status'])
    
        if response.json()['status'] == 'RUNNING':
            time.sleep(10)
            continue
        elif response.json()['status'] == 'QUEUED':
            print('Job queued, waiting...')
            time.sleep(15)
            continue
        else:
            if response.json()['status'] != "SUCCEEDED":
                print("Report did not SUCCEED: Job ID {} finished with status {}".format(report_ids[report_id], response.json()['status']))
            # Succeeds
            break

# Download reports and insert data into database
for (report_type, report_id) in report_ids.items():

    # download visit report
    response = requests.get('https://' + end_point + api + '/' + report_id + "/data", headers=headers)

    file_name = download_dir + report_type + '_' + today + '_' + file_suffix + '.csv'
    file = open(file_name, "w")
    file.write(response.text)
    file.close()

    # pass file to insert into DB.
    insert_csv_into_database(mydb, file_name)

print('Done.')    
